'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from typing import Optional, Dict, Any, Tuple

import pandas as pd

from pm4py.objects.ocel.obj import OCEL


def get_dataframes_from_ocel(
    ocel: OCEL, parameters: Optional[Dict[Any, Any]] = None
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    if parameters is None:
        parameters = {}

    events = ocel.events.copy()
    for col in events.columns:
        if str(events[col].dtype) == "object":
            events[col] = events[col].astype("string")
        elif "date" in str(events[col].dtype):
            events[col] = events[col].dt.strftime("%Y-%m-%dT%H:%M:%SZ")

    objects = ocel.objects.copy()
    for col in objects.columns:
        if str(objects[col].dtype) == "object":
            objects[col] = objects[col].astype("string")
        elif "date" in str(objects[col].dtype):
            objects[col] = objects[col].dt.strftime("%Y-%m-%dT%H:%M:%SZ")

    return events, objects
